/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright (c) 2024 NVIDIA Corporation & Affiliates
 */

#ifndef _RTE_FLOW_PERF_H_
#define _RTE_FLOW_PERF_H_

/**
 * @file
 * RTE generic flow perf API
 *
 * The generic flow perf library provides the ability to benchmark the performance
 * of flow based library, specifically speaking the RTE flow library.
 * This is done through exposing API that should be implemented by a specific
 * flow library so its performance can be evaluated.
 * The API is exposed through function pointers
 * that can be used within the generic flow perf library
 * implementation for evaluating performance.
 */

#include <stdint.h>

#include <rte_common.h>
#include <rte_ethdev.h>

#ifdef __cplusplus
extern "C" {
#endif

extern int rte_flow_perf_logtype;

#define RTE_FLOW_PERF_LOG(level, ...) \
	rte_log(RTE_LOG_ ## level, rte_flow_perf_logtype, "" __VA_ARGS__)

#define RTE_FLOW_PERF_MAX_PORT RTE_MAX_ETHPORTS
#define RTE_FLOW_PERF_BURST_TH 64

/**
 * @warning
 * @b EXPERIMENTAL: this structure may change without prior notice
 * Verbose error types.
 *
 * Most of them provide the type of the object referenced by struct
 * rte_flow_perf_error.cause.
 */
enum rte_flow_perf_error_type {
	RTE_FLOW_PERF_ERROR_TYPE_UNSPECIFIED,	/**< Cause unspecified. */
	RTE_FLOW_PERF_ERROR_TYPE_NO_MEM,	/**< Memory allocation error. */
	RTE_FLOW_PERF_ERROR_TYPE_PARSE,		/**< Parsing error. */
	RTE_FLOW_PERF_ERROR_TYPE_INVALID_CFG,	/**< Invalid configuration. */
};

/**
 * Verbose error structure definition.
 *
 * This object is normally allocated by applications and set by library, the
 * message points to a constant string which does not need to be freed by
 * the application, however its pointer can be considered valid only as long
 * as its associated DPDK port remains configured.
 *
 * Both cause and message may be NULL regardless of the error type.
 */
struct rte_flow_perf_error {
	enum rte_flow_perf_error_type type;	/**< Cause field and error types. */
	const void *cause;			/**< Object responsible for the error. */
	const char *message;			/**< Human-readable error message. */
};

/**
 * @warning
 * @b EXPERIMENTAL: this structure may change without prior notice
 *
 * This structure contains the sizes of all private information needed to be saved
 * by the library itself so it could allocate the pipeline configuration
 */
struct rte_flow_perf_alloc_cfg {
	/**< Flow perf layer port info private data size in bytes. */
	size_t port_info_priv_size;
	/**< Flow perf layer flow info private data size in bytes. */
	size_t flow_info_priv_size;
	/**< Flow perf layer number of flows info. */
	uint32_t nb_flow_info;
	/**< Flow perf layer item private data size in bytes. */
	size_t item_priv_size;
	/**< Flow perf layer idata private data size in bytes. */
	size_t *idata_priv_sizes;
	/**< Flow perf layer array of number of items one for each flow info. */
	uint32_t *nb_items;
	/**< Flow perf layer action private data size in bytes. */
	size_t action_priv_size;
	/**< Flow perf layer idata private data size in bytes. */
	size_t *adata_priv_sizes;
	/**< Flow perf layer array of number of actions one for each flow info */
	uint32_t *nb_actions;
	/**< Flow perf layer private data of struct rte_flow_perf_flow_data size in bytes */
	size_t *flow_data_priv_sizes;
	/**< Flow perf layer percore private data size in bytes for each flow. */
	size_t *core_data_priv_sizes;
	/**< Flow perf layer conf data size. */
	size_t conf_data_size;
};

/**
 * @warning
 * @b EXPERIMENTAL: this structure may change without prior notice
 *
 * Enumeration defining the traffic handling mode. Defines whether
 * to run the library to benchmark background flows or only for
 * measuring deletion/insertion rate
 */
enum rte_flow_perf_traffic_mode {
	RTE_FLOW_PERF_TG_NONE,
	RTE_FLOW_PERF_TG_FWD,
	RTE_FLOW_PERF_TG_FWD_BG,
};

/**
 * @warning
 * @b EXPERIMENTAL: this enum may change without prior notice
 *
 * Defines type of object (flow item or action) to be updated.
 */
enum rte_flow_perf_conf_type {
	RTE_FLOW_PERF_CONF_ITEM = 0,
	RTE_FLOW_PERF_CONF_ACTION = 1,
};

/**
 * @warning
 * @b EXPERIMENTAL: this enum may change without prior notice.
 *
 * Flows operation insert or delete.
 */
enum rte_flow_perf_mode {
	RTE_FLOW_PERF_ADD,
	RTE_FLOW_PERF_DEL,
	RTE_FLOW_PERF_MAX,
};

/**
 * @warning
 * @b EXPERIMENTAL: this enum may change without prior notice.
 *
 * IPv4 field update flags.
 */
enum PERF_ITEM_IPV4_FIELD_FLAG {
	RTE_FLOW_PERF_ITEM_IPV4_SRC = 1 << 0,
	RTE_FLOW_PERF_ITEM_IPV4_DST = 1 << 1,
	RTE_FLOW_PERF_ITEM_TRANSPORT_SRC = 1 << 2,
	RTE_FLOW_PERF_ITEM_TRANSPORT_DST = 1 << 3,
	RTE_FLOW_PERF_VXLAN_TUN_ID = 1 << 4,
};

/**
 * @warning
 * @b EXPERIMENTAL: this enum may change without prior notice.
 *
 * IPv6 field update flags.
 */
enum RTE_FLOW_PERF_ITEM_IPV6_FIELD_FLAG {
	RTE_FLOW_PERF_ITEM_IPV6_HOP_LIMITS = 1 << 0,
};

/**
 * @warning
 * @b EXPERIMENTAL: this enum may change without prior notice.
 *
 * TCP field update flags.
 */
enum RTE_FLOW_PERF_ITEM_TCP_FIELD_FLAG {
	RTE_FLOW_PERF_ITEM_TCP_FLAGS = 1 << 0,
};

/**
 * @warning
 * @b EXPERIMENTAL: this enum may change without prior notice.
 *
 * Meta data match field update flags.
 */
enum RTE_FLOW_PERF_ITEM_META_FIELD_FLAG {
	RTE_FLOW_PERF_ITEM_META = 1 << 0,
	RTE_FLOW_PERF_ITEM_TAG = 1 << 1,
};

/**
 * @warning
 * @b EXPERIMENTAL: this enum may change without prior notice.
 *
 * Meta data action field update flags.
 */
enum RTE_FLOW_PERF_ACTION_META_FIELD_FLAG {
	RTE_FLOW_PERF_ACTION_SET_META = 1 << 0,
	RTE_FLOW_PERF_ACTION_SET_TAG = 1 << 1,
};

/**
 * @warning
 * @b EXPERIMENTAL: this enum may change without prior notice.
 *
 * JUMP action field update flags.
 */
enum RTE_FLOW_PERF_ACTION_JUMP_FIELD_FLAG {
	RTE_FLOW_PERF_ACTION_JUMP = 1 << 0,
};

/**
 * @warning
 * @b EXPERIMENTAL: this enum may change without prior notice.
 *
 * Items or actions field update mode.
 */
enum RTE_FLOW_PERF_FIELD_UPDATE_MODE {
	RTE_FLOW_PERF_FIELD_UPDATE_NONE,
	RTE_FLOW_PERF_FIELD_UPDATE_INC,
	RTE_FLOW_PERF_FIELD_UPDATE_DEC,
	RTE_FLOW_PERF_FIELD_UPDATE_CB,
};

/**
 * @warning
 * @b EXPERIMENTAL: this enum may change without prior notice.
 *
 * Port forwarding mode.
 */
enum RTE_FLOW_PERF_PORT_FWD_FLAG {
	RTE_FLOW_PERF_PORT_FWD_RX_ONLY,
	RTE_FLOW_PERF_PORT_FWD_RX_SELF_TX_SELF,
	RTE_FLOW_PERF_PORT_FWD_RX_SELF_TX_OTHER,
};

/**
 * @warning
 * @b EXPERIMENTAL: this structure may change without prior notice.
 *
 * Represents an item to be updated in the pipeline
 */
struct rte_flow_perf_idata {
	/**< Number of items */
	uint32_t dyn_item_num;
	/**< The configuration info */
	struct rte_flow_perf_conf_info *conf;
	/**< The configuration object */
	struct rte_flow_perf_conf_obj *obj;
	/**< Size item's private data */
	size_t idata_priv_size;
	/**< Item's private data, managed by implementing application */
	uint8_t *idata_priv;
};

/**
 * @warning
 * @b EXPERIMENTAL: this structure may change without prior notice.
 *
 * Represent an action to be updated in the pipeline
 */
struct rte_flow_perf_adata {
	/**< Number of actions */
	uint32_t dyn_act_num;
	/**< The configuration info */
	struct rte_flow_perf_conf_info *conf;
	/**< The configuration object */
	struct rte_flow_perf_conf_obj *obj;
	/**< Size of action's private data. */
	size_t adata_priv_size;
	/**< Action's private data, used by implementing application. */
	uint8_t *adata_priv;
};

/**
 * @warning
 * @b EXPERIMENTAL: this structure may change without prior notice.
 *
 * Contains statistics gathered during performance measurements.
 */
struct rte_flow_perf_stat {
	/**< Array of number of CPU cycles needed for inserting and deleting. */
	uint64_t cycles[RTE_FLOW_PERF_MAX];
	/**< number of flows that failed to be inserted */
	uint32_t miss_count;
	/**< Array of pushed/pending flow one per port */
	struct {
		/**< Allocated/destroyed flow. */
		uint32_t pushed_flow;
		/**< Pending to pull. */
		uint32_t pending_flow;
	} pst[RTE_FLOW_PERF_MAX_PORT];
} __rte_cache_aligned;

/**
 * @warning
 * @b EXPERIMENTAL: this structure may change without prior notice.
 *
 * Represent the indirect actions to be used. For now it is highly coupled
 * with RTE flow library, however this will change and will be represented in
 * more generic way.
 */
struct perf_flow_indir_action {
	LIST_ENTRY(perf_flow_indir_action) next;
	uint32_t port_id;
	uint32_t total;
	uint32_t cur;
	/* The action to be updated. */
	struct rte_flow_action *action;
	struct rte_flow_action_handle **ia;
};

/**
 * @warning
 * @b EXPERIMENTAL: this structure may change without prior notice.
 *
 * Represents an object (action or item) to be updated
 */
struct rte_flow_perf_conf_data {
	/**< Overall size of the configuration private data */
	size_t *conf_data_size;
	/**< the configuration data (action or item) to be updated */
	uint8_t *conf_data;
};

/**
 * @warning
 * @b EXPERIMENTAL: this API may change without prior notice.
 *
 *  Callback used for updating an action or item in the pipeline
 *  when inserting/deleting flows.
 *
 * @param[in] obj
 *   Object (action or item) configuration to indicate how to update the object
 * @param[out] error
 *   Perform verbose error reporting if not NULL.
 */
typedef void (*perf_conf_update_t)(struct rte_flow_perf_conf_obj *obj);

/**
 * @warning
 * @b EXPERIMENTAL: this structure may change without prior notice.
 *
 * Represents how to update a single element (action or item)
 */
struct rte_flow_perf_element {
	/**< Item update mode. */
	uint32_t mode;
	/**< Fields to be updated. */
	uint32_t field_flags;
	union {
		/**< Update by callback. */
		perf_conf_update_t cb;
		struct {
			/**< Element update steps. Reserved. */
			uint32_t steps;
			/**< Max value. */
			uint32_t max_mask;
		};
	};
	/**< Flow perf layer element private data size in bytes*/
	size_t element_priv_size;
	/**< Flow perf layer element private data */
	uint8_t *element_priv;
};

/**
 * @warning
 * @b EXPERIMENTAL: this structure may change without prior notice.
 *
 * Represent how an object should be updated.
 */
struct rte_flow_perf_conf_info {
	/**< indicates the type of updating an object */
	uint32_t type;
	union {
		/**< item info to be updated. */
		struct rte_flow_perf_element *pfi;
		/**< action info to be updated. */
		struct rte_flow_perf_element *pfa;
	};
	/**< a callback to be used for updating the object if the type field indicates so */
	perf_conf_update_t cb;
};

/**
 * @warning
 * @b EXPERIMENTAL: this structure may change without prior notice.
 *
 * Represent the object to be updated and how it should be updated.
 */
struct rte_flow_perf_conf_obj {
	struct rte_flow_perf_conf_info *info;		/**< Flow update info. */
	struct rte_flow_perf_conf_data data[2];		/**< Item & action update data. */
};

/**
 * @warning
 * @b EXPERIMENTAL: this structure may change without prior notice.
 *
 * Contains all data needed for flows insertion/deletion per core
 */
struct rte_flow_perf_per_core_data {
	union {
		/**< Flow item or action initial value. */
		uint32_t start;
		/**< Allocated/destroyed flow. */
		uint32_t pushed;
	};
	/**< Number of overall successful operation. */
	uint32_t count;
	/**< pointer to the flow data array*/
	struct rte_flow_perf_flow_data *pfd;
	/**< pointer to the action data array*/
	struct rte_flow_perf_adata **pfa;
	/**< pointer to the item data array */
	struct rte_flow_perf_idata **pfi;
	/**< Configuration object to extract how an object should be updated */
	struct rte_flow_perf_conf_obj *obj;
	/**< Per core private data for storing per core data */
	size_t core_data_priv_size;
	/**< Per core private data size */
	uint8_t *core_data_priv;
};

/**
 * @warning
 * @b EXPERIMENTAL: this structure may change without prior notice.
 *
 * Contains all data needed for one pipe flows insertion/deletion
 */
struct rte_flow_perf_flow_info {
	/**< Total flow number */
	uint32_t nb_flows;
	/**< Static background flows to be created in advanced. */
	uint32_t nb_background_flows;
	/**< Flow index in the pipe  */
	bool index_flow;
	/**< Measure insertion rate or not. */
	bool measure;
	/**< Reserved, will impelemnt in next phase. */
	uint32_t batch_size;
	/**< Number of item element list. */
	uint32_t nb_pfi;
	/**< Item element list. */
	struct rte_flow_perf_element **items;
	/**< Number of action element list. */
	uint32_t nb_pfa;
	/**< Action element list. */
	struct rte_flow_perf_element **actions;
	/**< Flow perf layer flow info private data size in bytes*/
	size_t flow_info_priv_size;
	/**< Flow perf layer flow info private data */
	uint8_t *flow_info_priv;
};

/**
 * @warning
 * @b EXPERIMENTAL: this structure may change without prior notice.
 *
 * Contains all data needed for port initialization and to prepare it for
 * the different measurements.
 */
struct rte_flow_perf_port_info {
	/**< Port index. */
	uint32_t port_id;
	/**< Number of HWS queues. */
	uint32_t nb_flow_queue;
	/**< Size of HWS queues. */
	uint32_t queue_size;
	/**< Number of pipes/flow_info in the pipeline. */
	uint32_t nb_flow_info;
	/**< Flow info array; array representing the pipeline */
	struct rte_flow_perf_flow_info *flow_info;
	/**< Flow perf layer port info private data size in bytes */
	size_t port_info_priv_size;
	/**< Flow perf layer port info private data */
	uint8_t *port_info_priv;
};

/**
 * @warning
 * @b EXPERIMENTAL: this structure may change without prior notice.
 *
 * Contains all data needed for a pipe creation and data needed for flows isertion
 * or deletion to the same pipe.b This structure is used to represent a pipe
 * in the pipeline, hence it has the "next" field.
 */
struct rte_flow_perf_flow_data {
	/**< next pipe data or flow data */
	LIST_ENTRY(rte_flow_perf_flow_data) next;
	/**< indirect action list for the pipe */
	LIST_HEAD(ialist_t, perf_flow_indir_action) indir_action_list;
	/**< mtr profile id */
	uint32_t mtr_profile_id;
	/**< number of flow perf element items */
	uint32_t nb_pfi;
	/**< number of flow perf element actions */
	uint32_t nb_pfa;
	/**< number of background flows in the pipe */
	uint32_t nb_bg_flows;
	/**< number of dynamic flows in the pipe */
	uint32_t nb_dyn_flows;
	/**< number of cores used for the flows operations */
	uint32_t nb_pcore_flows;
	/**< number of item elements to be updated */
	uint32_t pconf_num;
	/**< Maximum number of flows to be inserted/deleted in this pipe */
	uint32_t mux;
	/**< The index of the current flow */
	bool index_flow;
	/**< whether or not to measure the bandwidth */
	bool measure;
	/**< Flow info, pipe needed info */
	struct rte_flow_perf_flow_info *fi;
	/**< Port info, port needed info to insert/delete the flows to/from */
	struct rte_flow_perf_port_info *port_info;
	/**< per core data for flows handling */
	struct rte_flow_perf_per_core_data **pcfd;
	/**< Actions data update configuration object */
	struct rte_flow_perf_adata **pfa;
	/**< Items data update configuration object */
	struct rte_flow_perf_idata **pfi;
	/**< Private data such as the created pipe/table */
	size_t flow_data_priv_size;
	/**< Private data size */
	uint8_t *flow_data_priv;
};

/**
 * @warning
 * @b EXPERIMENTAL: this structure may change without prior notice.
 *
 * The perf flow manager contains the pipeline info data
 */
struct rte_flow_perf_mng {
	/**< The background flows pipes data */
	LIST_HEAD(flow_blist_t, rte_flow_perf_flow_data) bg_flow_list;
	/**< The dynamic flows (to be inserted/deleted) pipes data */
	LIST_HEAD(flow_dlist_t, rte_flow_perf_flow_data) dyn_flow_list;
	/**< Number of pipes in the dynamic list */
	uint32_t nb_flow;
	/**< Number of HWS queues */
	uint32_t nb_queue;
	/**< Number of iterations of flow insertion or deletion */
	uint32_t loop;
	/**< Number of total flows */
	uint32_t pcore_total_flows;
	/**< The results of insertion/deletion rate */
	struct rte_flow_perf_stat *stat;
} __rte_cache_aligned;

/**
 * @warning
 * @b EXPERIMENTAL: this structure may change without prior notice.
 *
 * The pipeline contains all needed data as parsed from the provided JSON file
 * so it could be used for  configuring the flow perf manager.
 * Therefore, this structure should be used by the users while parsing, only after
 * allocating the memory needed for this struct
 * by calling rte_flow_perf_allocate_pipeline
 */
struct rte_flow_perf_pipeline {
	/**< Pipeline name. */
	char name[256];
	/**< Port info number. */
	uint32_t nb_port_info;
	/**< Port Number of RX queues. */
	uint16_t nb_rxq;
	/**< Port number of TX queues. */
	uint16_t nb_txq;
	/**< Port number of hairpin queues. */
	uint16_t nb_hairpinq;
	/**< Port hairpin configuration */
	uint32_t hairpin_conf;
	/**< Traffic handling mode. */
	uint32_t traffic_mode;
	/**< Packet forwarding mode. */
	uint32_t fwd_mode;
	/**< Flow insert/delete loop. */
	uint32_t loop;
	/**< Port info. */
	struct rte_flow_perf_port_info *port_info;
};

/**
 * @warning
 * @b EXPERIMENTAL: this structure may change without prior notice
 *
 * The structure contains APIs to be used for performance benchmarking, such as a concrete
 * implementation of flows insertion/deletion using RTE flow library.
 */
struct rte_flow_perf_api {
	/**< Function pointer for parsing the pipe configuration */
	struct rte_flow_perf_pipeline * (*parse)(const char *json_file_path);
	/**< Function pointer for initializing the concrete flow perf layer. */
	int (*init)(struct rte_flow_perf_pipeline *ppl);
	/**< Function pointer for creating the pipes. */
	int (*create_pipes)(struct rte_flow_perf_flow_info *fi,
					struct rte_flow_perf_flow_data *pfd);
	/**< Function pointer for packet handling and BW measuring if needed. */
	int (*packet_handling)(void);
	/**< Function pointer for destruction of the concrete flow perf layer. */
	int (*destroy)(void);
	/**< Function pointer for updating field addresses in idata private data */
	void (*set_idata_priv_field_addr)
			(struct rte_flow_perf_idata *pfit);
	/**< Function pointer for updating field addresses of adata private data */
	void (*set_adata_priv_field_addr)
			(struct rte_flow_perf_adata *pfat, uint32_t anum);
	/**< Function pointer for updating field values of idata private data */
	void (*set_idata_priv_field_vals)
		(struct rte_flow_perf_element *pfi, uint32_t nb_items, bool init_v,
			struct rte_flow_perf_idata *pfit);
	/**< Function pointer for updating field values of adata private data */
	void (*set_adata_priv_field_vals)
	(struct rte_flow_perf_flow_data *pfd, struct rte_flow_perf_element *pat,
		uint32_t nb_actions, bool init_v, uint32_t nb_flow, bool clone,
		struct rte_flow_perf_adata *pfat);
	/**< Function pointer for adding background flows */
	int (*add_bg_flows)
		(struct rte_flow_perf_flow_data *pfd, uint32_t postpone,
			uint32_t i, uint32_t *fidx, uint32_t *miss_count);
	/**< Function pointer for removing background flows */
	int (*rm_bg_flows)
		(struct rte_flow_perf_flow_data *pfd, uint32_t postpone,
			uint32_t i, uint32_t *fidx, uint32_t *miss_count);
	/**< Function pointer for handling dynamic flows */
	int (*handle_dyn_flows)
	(uint16_t port_id, uint16_t queue,
		enum rte_flow_perf_mode mode,
		struct rte_flow_perf_stat *st,
		struct rte_flow_perf_flow_data *pfd,
		struct rte_flow_perf_per_core_data *pcfd);
	/**< Function pointer for pushing dynamic flows operations */
	int (*push_dyn_flows)
	(uint32_t queue, uint32_t *fidx, struct rte_flow_perf_stat *st);
	/**< Function pointer for destroying flow data private allocated fields */
	void (*destroy_flow_data_priv)
	(uint32_t port_id, void *flow_data_priv);
	/**< Function pointer for destroying idata private allocated fields */
	void (*destroy_idata_priv)
	(uint32_t port_id, void *idata_priv);
	/**< Function pointer for destroying adata private allocated fields */
	void (*destroy_adata_priv)
	(uint32_t port_id, void *adata_priv);
};

/**
 * @warning
 * @b EXPERIMENTAL: this API may change without prior notice.
 *
 * Allocates the memory for the pipeline so it could be used by
 * the JSON parser in the flow perf layer in order to configure the pipeline.
 * Therefore, users should call this API from within rte_flow_perf_layer_parse
 * once the needed information is parsed, such as number of pipes (flows info).
 *
 * @param[in] alloc_cfg
 *    the actual needed allocation information needed for the pipeline allocation
 * @param[out] error
 *   Perform verbose error reporting if not NULL.
 * @return
 *   A pointer to a new allocated pipeline struct on success, NULL otherwise.
 */
__rte_experimental
struct rte_flow_perf_pipeline *
rte_flow_perf_allocate_pipeline(struct rte_flow_perf_alloc_cfg *alloc_cfg,
			struct rte_flow_perf_error *error);

/**
 * @warning
 * @b EXPERIMENTAL: this API may change without prior notice.
 *
 * Executes the flow performance benchmarking APIs in the following manner:
 *	1.Set the flow perf layer concrete implementation
 *	2.Parses the arguments and parameters
 *	3.Initializes resources and create pipes
 *	4.Launch all lcores for flows insertion/deletion and packet handling.
 *	5.Wait for all lcores to finish and dump the results
 *	6.Destroy all allocated resources
 *
 * @param[in] cfg
 *    APIs to be used for parsing and flows handling.
 * @param[in] json_file_path
 *    JSON file path for parsing pipeline configuration.
 * @param[out] error
 *   Perform verbose error reporting if not NULL.
 * @return
 *   0 on success, a negative value otherwise.
 */
__rte_experimental
int
rte_flow_perf_execute(struct rte_flow_perf_api *cfg, const char *json_file_path,
		struct rte_flow_perf_error *error);

/**
 * @warning
 * @b EXPERIMENTAL: this API may change without prior notice.
 *
 * Sets flow perf default API implementation
 *
 * @param[out] cfg
 *    flow perf default function pointer implementation.
 */
__rte_experimental
void
rte_flow_perf_set_config(struct rte_flow_perf_api *cfg);

#ifdef __cplusplus
}
#endif

#endif /* _RTE_FLOW_PERF_H_ */
